(in-package :iolib.examples)

(defparameter *event-base* nil)
;; This holds any open connections to clients as keys in the table. The values
;; is a list containing the host and port of the connection. We use this to
;; close all connections to the clients, if any, when the server exits.  This
;; allows all clients to notice the server had gone away.
(defparameter *open-connections* nil)

;; This program uses the multiplexer to select on a listening socket which
;; when it becomes ready in the read bit, will do a blocking accept, and then
;; register a callback then when it is ready will do a blocking read and
;; then immediately a blocking write back to the same socket.
;; This is not multithreaded.

(defun run-ex6-server-helper (port)
  (with-open-socket
      (server :connect :passive
              :address-family :internet
              :type :stream
              :ipv6 nil
              :external-format '(:utf-8 :eol-style :crlf))
    (format t "Created socket: ~A[fd=~A]~%"
            server (socket-os-fd server))

    ;; Bind the socket to all interfaces with specified port.
    (bind-address server +ipv4-unspecified+
                  :port port
                  :reuse-addr t)
    (format t "Bound socket: ~A~%" server)

    ;; start listening on the server socket
    (listen-on server :backlog 5)
    (format t "Listening on socket bound to: ~A:~A~%"
            (local-name server)
            (local-port server))

    ;; Set up the initial listener handler for any incoming clients
    ;; What kind of error checking do I need to do here?
    (set-io-handler *event-base*
                    (socket-os-fd server)
                    :read
                    (make-ex6-server-listener-handler server))

    ;; keep accepting connections forever.
    (handler-case
        (event-dispatch *event-base*)

      (hangup ()
        (format t "Caught unexpected hangup! Cilent closed connection on write!"))
      (end-of-file ()
        (format t "Caught unexpected end-of-file! Client closed connection on read!")))))

(defun make-ex6-server-listener-handler (socket)
  (lambda (fd event exception)
    ;; do a blocking accept, returning nil if no socket
    (let* ((client (accept-connection socket :wait t))
           (who (remote-name client))
           (port (remote-port client)))
      (when client
        (format t "Accepted a client from ~A:~A~%" who port)

	
        ;; save the client connection in case we need to close it later.
        (setf (gethash `(,who ,port) *open-connections*) client)

        ;; set up an line echo function for the client socket.
        (set-io-handler *event-base*
                        (socket-os-fd client)
                        :read
                        (make-ex6-server-line-echoer
                         client
                         who
                         port
                         (make-ex6-server-disconnector client)))))))

;; This function returns a function that reads a line, then echoes it
;; right back onto the socket it came from. This is blocking i/o.
;; This code can suffer denial of service attacks like on page 167.
(defun make-ex6-server-line-echoer (socket who port disconnector)
  (format t "Creating line-echoer for ~A:~A~%" who port)
  (lambda (fd event exception)
    (handler-case
        (let ((line (read-line socket))) ;; read a line from the client
          (format t "Read ~A:~A: ~A~%" who port line)
          (format socket "~A~%" line) ;; write it the client
          (finish-output socket)
          (format t "Wrote ~A:~A: ~A~%" who port line)

          (when (string= line "quit")
            (format t "Client requested quit!~%")
            (funcall disconnector who port)))

      (hangup ()
        (format t "Client went away on a write.~%")
        (funcall disconnector who port))
      (end-of-file ()
        (format t "Client went away on a read.~%")
        (funcall disconnector who port)))))

(defun make-ex6-server-disconnector (socket)
  (lambda (who port)
    (format t "Closing connection to ~A:~A~%" who port)
    (remove-fd-handlers *event-base* (socket-os-fd socket))
    (close socket)
    (remhash `(,who ,port) *open-connections*)))


;; This just checks for some error conditions so we can print out a nice
;; message about it.
(defun run-ex6-server (&key (port *port*))
  (unwind-protect
       (handler-case
           (progn
             (setf *open-connections* (make-hash-table :test #'equalp))
             (setf *event-base* (make-instance 'event-base))
             (run-ex6-server-helper port))
         ;; handle some common signals
         (socket-address-in-use-error ()
           (format t "Bind: Address already in use, forget :reuse-addr t?")))


    ;; close all open connections to the clients, if any
    (maphash
     #'(lambda (k v)
         (format t "Closing a client connection to ~A~%" k)
         (close v))
     *open-connections*)

    ;; and clean up the event-base too!
    (close *event-base*)))