(in-package :iolib.examples)

(defparameter *event-base* nil)

;; This program differs from ex6a-client in the fact that when the end-of-file
;; is reached for the input, only the write end of the socket is shutdown, this
;; allows inflight data to reach the server.

(defun run-ex6b-client-helper (port)

  ;; Create a internet TCP socket under IPV4
  (with-open-socket
      (socket :connect :active
              :address-family :internet
              :type :stream
              :external-format '(:utf-8 :eol-style :crlf)
              :ipv6 nil)

    ;; do a blocking connect to the echo server on the port.
    (connect socket (lookup-hostname "localhost") :port port :wait t)

    (format t "Connected to server ~A:~A from my local connection at ~A:~A!~%"
            (remote-name socket) (remote-port socket)
            (local-name socket) (local-port socket))

    ;; set up the handlers for read and write
    (set-io-handler *event-base* (socket-os-fd socket)
                    :read
                    (make-ex6b-str-cli-read
                     socket
                     (make-ex6b-client-disconnector socket)))

    (set-io-handler *event-base* (socket-os-fd socket)
                    :write
                    (make-ex6b-str-cli-write
                     socket
                     (make-ex6b-client-disconnector socket)))

    (handler-case
        ;; keep processing input and output on the fd by calling he relevant
        ;; handlers as the socketbecome ready.
        (event-dispatch *event-base*)

      (hangup ()
        (format t "Got uncaught hangup. Server closed connection on write!!%"))
      (end-of-file ()
        (format t "Got uncaught end-of-file. Server closed connection on read!!%")))))

(defun make-ex6b-str-cli-write (socket disconnector)
  ;; When this next function gets called it is because the event dispatcher
  ;; knows the socket to the server is writable.
  (lambda (fd event exception)
    ;; Get a line from stdin, and send it to the server
    (handler-case
        (let ((line (read-line)))
          (format socket "~A~%" line)
          (finish-output socket))
      (end-of-file ()
        (format t
                "make-ex6b-str-cli-write: User performed end-of-file!~%")
        ;; Shutdown the write end of my pipe to give the inflight data the
        ;; ability to reach the server!
        (format t
                "make-ex6b-str-cli-write: Shutting down write end of socket!~%")
        (shutdown socket :write t)
        ;; just remove the write handler, since I won't need it anymore!
        (funcall disconnector :write))
      (hangup ()
        (format t
                "make-ex6b-str-cli-write: server closed connection on write!~%")
        (funcall disconnector)))))

(defun make-ex6b-str-cli-read (socket disconnector)
  ;; When this next function gets called it is because the event dispatcher
  ;; knows the socket from the server is readable.
  (lambda (fd event exception)
    ;; get a line from the server, and send it to *standard-output*
    (handler-case
        (let ((line (read-line socket)))
          (format t "~A~%" line)
          (finish-output))
      (end-of-file ()
        (format t "make-ex6b-str-cli-read: server closed connection on read!~%")
        (funcall disconnector)))))

(defun make-ex6b-client-disconnector (socket)
  ;; When this function is called, it can be told which callback to remove, if
  ;; no callbacks are specified, all of them are removed!
  (lambda (&rest events)
    (format t "Disconnecting socket ~A~%" socket)
    (let ((fd (socket-os-fd socket)))
      (if (null events)
          (remove-fd-handlers *event-base* fd :read t :write t :error t)
          (progn
            (when (member :read events)
              (remove-fd-handlers *event-base* fd :read t))
            (when (member :write events)
              (remove-fd-handlers *event-base* fd :write t))
            (when (member :error events)
              (remove-fd-handlers *event-base* fd :error t))))
      ;; I don't close the socket here because this socket is the exact same
      ;; socket as in the with-open-socket call. In this program, when one
      ;; callback closes the connection both handlers are removed, which
      ;; makes the event-base empty, which makes it return, which falls past
      ;; the end of the with-open-socket, and thus the socket is closed.
      #|(close socket)|#)))

(defun run-ex6b-client (&key (port *port*))
  (unwind-protect
       (progn
         ;; When the connection gets closed, either intentionally in the client
         ;; or because the server went away, we want to leave the multiplexer
         ;; event loop. So, when making the event-base, we explicitly state
         ;; that we'd like that behavior.
         (setf *event-base*
               (make-instance 'iomux:event-base :exit-when-empty t))
         (handler-case
             (run-ex6b-client-helper port)
           ;; handle a commonly signaled error...
           (socket-connection-refused-error
               ()
             (format t "Connection refused to ~A:~A. Maybe the server isn't running?~%"
                     (lookup-hostname "localhost") port)))))

  ;; ensure we clean up the event base regardless of how we left the client
  ;; algorithm
  (close *event-base*)
  (format t "Client Exited.~%"))




